package com.ibm.oti.shared;

import com.ibm.oti.util.Util;
import java.io.File;
import java.net.MalformedURLException;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URI;
import java.security.AccessController;
import java.security.PrivilegedAction;

/*
 * Licensed Materials - Property of IBM,
 *     Copyright IBM Corp. 1998, 2010  All Rights Reserved
 */

/**
 * Abstract SharedClassHelper provides common functions and data to class helper subclasses.
 * <p>
 * @author		OTI
 * @version		initial
 *
 * @see SharedClassHelper
 * @see SharedAbstractHelper
 */
public abstract class SharedClassAbstractHelper extends SharedAbstractHelper implements SharedClassHelper {

	int ROMCLASS_COOKIE_SIZE;
	private static final int URI_EXCEPTION = 1;
	private static final int FILE_EXIST = 2;
	private static final int FILE_NOT_EXIST = 3;
	private SharedClassFilter sharedClassFilter;

	static byte[] nativeFlags = new byte[1];
	static final int CACHE_FULL_FLAG = 0;

	static {
		nativeFlags[CACHE_FULL_FLAG] = 0;
	}

	/**
	 * Utility function. Determines whether a byte array being passed to defineClass is a class found
	 * in the shared class cache, or a class found locally.
	 * <p>
	 * @param 		classBytes a potential shared class cookie
	 * @return 		true if bytes are a cookie.
	 */
	public boolean isSharedClassCookie(byte[] classBytes) {
		return (classBytes.length == ROMCLASS_COOKIE_SIZE);
	}

	private native int initializeShareableClassloaderImpl(ClassLoader loader);

	void initializeShareableClassloader(ClassLoader loader) {
		/* Allow ClassLoader to be Garbage Collected by keeping a WeakReference.
		 * This is important as any live references to the ClassLoader will
		 * prevent it being collected. CMVC 98943. */
		ROMCLASS_COOKIE_SIZE = initializeShareableClassloaderImpl(loader);
	}

	URL convertJarURL(URL url) {
		if (url==null)
			return null;

		if (!url.getProtocol().equals("jar")) {
			return url;
		} else {
			String jarPath = recursiveJarTrim(url.getPath());
			try {
				return new URL(jarPath);
			} catch (MalformedURLException e) {
				return url;
			}
		}
	}

	private String recursiveJarTrim(String jarName) {
		if (jarName==null)
			return null;

		int lastIndexBang = jarName.lastIndexOf("!/");
		boolean startsWithJar = jarName.startsWith("jar:");
		int subStringStart = startsWithJar ? 4 : 0;
		int subStringEnd = lastIndexBang>=0 ? lastIndexBang : jarName.length();

		if (!startsWithJar && lastIndexBang==-1) {
			return jarName;
		} else {
			return recursiveJarTrim(jarName.substring(subStringStart, subStringEnd));
		}
	}

	boolean validateClassLoader(ClassLoader loader, Class clazz) {
		if (loader==null) {
			return false;
		}
		if (!loader.equals(clazz.getClassLoader())) {
			printVerboseError("ClassLoader of SharedClassHelper and ClassLoader of class "+clazz.getName()+" do not match.");
			return false;
		}
		return true;
	}

	boolean validateURL(URL url, boolean checkExists) {
		String protocol = url.getProtocol();
		if (!(protocol.equals("file") || protocol.equals("jar"))) {
			printVerboseInfo("URL "+url+" does not have required file or jar protocol.");
			return false;
		}
		if (checkExists) {
			Integer fExists;
			final URL urlToCheck = url;
			fExists = (Integer)AccessController.doPrivileged(new PrivilegedAction() {
				public Object run() {
					File f;
					try {
						f = new File(urlToCheck.toURI());
					} catch (URISyntaxException e) {
						// try URL encoding the path
						try {
							f = new File(new URI(Util.urlEncode(urlToCheck.toExternalForm())));
						} catch (URISyntaxException e2) {
							// try the simple string path
							f = new File(urlToCheck.getPath());
						} catch (IllegalArgumentException e2) {
							// as below
							f = new File(urlToCheck.getPath());
						}
					} catch (IllegalArgumentException e) {
						/* There is a bug in the URI code which does not handle all UNC paths correctly
						 * If we hit this, revert to using simple string path */
						f = new File(urlToCheck.getPath());
					}
					return new Integer((f.exists() ? FILE_EXIST : FILE_NOT_EXIST));
				}
			});
			if (fExists.intValue() == FILE_NOT_EXIST) {
				printVerboseError("URL resource "+url+" does not exist.");
				return false;
			} else
			if (fExists.intValue() == URI_EXCEPTION) {
				printVerboseError("URI could not be created from the URL "+url);
				return false;
			}
		}
		return true;
	}

	/**
	 * Set the sharedClassFilter for a helper instance.  Supplying
	 * null will remove any filter currently associated with the
	 * helper instance.
	 * <p>
	 * @param filter The filter to use when finding and storing classes.
	 */
	public synchronized void setSharingFilter(SharedClassFilter filter) {
		if (System.getSecurityManager()!=null) {
			if (!checkReadPermission()) {
				printVerboseError("Read permission denied. Cannot set sharing filter.");
				return;
			}
			if (!checkWritePermission()) {
				printVerboseError("Write permission denied. Cannot set sharing filter.");
				return;
			}
		}
		this.sharedClassFilter = filter;
	}

	/**
	 * Return the sharedClassFilter associated with this helper.
	 * <p>
	 * @return the filter instance or null if non is associated
	 */
	public synchronized SharedClassFilter getSharingFilter() {
		return this.sharedClassFilter;
	}
}
